"""This module contains all the definitions for yMMSL."""
from collections import OrderedDict
from typing import Any, List, Optional, Union, cast
from typing import Dict     # noqa

import yatiml

from ymmsl.component import Operator    # noqa
from ymmsl.component import Component
from ymmsl.identity import Identifier, Reference


class Conduit:
    """A conduit transports data between simulation components.

    A conduit has two endpoints, which are references to a port on a
    simulation component. These references must be of one of the
    following forms:

    - component.port
    - namespace.component.port (or several namespace prefixes)

    Attributes:
        sender: The sending port that this conduit is connected to.
        receiver: The receiving port that this conduit is connected to.

    """

    def __init__(self, sender: str, receiver: str) -> None:
        """Create a Conduit.

        Args:
            sender: The sending component and port, as a Reference.
            receiver: The receiving component and port, as a
                    Reference.

        """
        self.sender = Reference(sender)
        self.receiver = Reference(receiver)

        self.__check_reference(self.sender)
        self.__check_reference(self.receiver)

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return 'Conduit({} -> {})'.format(self.sender, self.receiver)

    def __eq__(self, other: Any) -> bool:
        """Returns whether the conduits connect the same ports."""
        if not isinstance(other, Conduit):
            return NotImplemented
        return self.sender == other.sender and self.receiver == other.receiver

    @staticmethod
    def __check_reference(ref: Reference) -> None:
        """Checks an endpoint for validity."""
        # check that subscripts are at the end
        for i, part in enumerate(ref):
            if isinstance(part, int):
                if (i+1) < len(ref) and isinstance(ref[i+1], Identifier):
                    raise ValueError('Reference {} contains a subscript that'
                                     ' is not at the end, which is not allowed'
                                     ' in conduits.'.format(ref))

        # check that the length is at least 2
        if len(Conduit.__stem(ref)) < 2:
            raise ValueError((
                    'Senders and receivers in conduits must have a component'
                    ' name, a period, and then a port name and optionally a'
                    ' slot. Reference {} is missing either the component or'
                    ' the port. Did you perhaps type a comma or an underscore'
                    ' instead of a period? It should be "component.port"'
                    ).format(ref))

    def sending_component(self) -> Reference:
        """Returns a reference to the sending component."""
        return cast(Reference, self.__stem(self.sender)[:-1])

    def sending_port(self) -> Identifier:
        """Returns the identity of the sending port."""
        # We've checked that it's an Identifier during construction
        return cast(Identifier, self.__stem(self.sender)[-1])

    def sending_slot(self) -> List[int]:
        """Returns the slot on the sending port.

        If no slot was given, an empty list is returned.

        Returns:
            A list of slot indexes.

        """
        return self.__slot(self.sender)

    def receiving_component(self) -> Reference:
        """Returns a reference to the receiving component."""
        return cast(Reference, self.__stem(self.receiver)[:-1])

    def receiving_port(self) -> Identifier:
        """Returns the identity of the receiving port."""
        return cast(Identifier, self.__stem(self.receiver)[-1])

    def receiving_slot(self) -> List[int]:
        """Returns the slot on the receiving port.

        If no slot was given, an empty list is returned.

        Returns:
            A list of slot indexes.

        """
        return self.__slot(self.receiver)

    @staticmethod
    def __slot(reference: Reference) -> List[int]:
        """Extracts the slot from the given reference.

        The slot is the list of contiguous ints at the end of the
        reference. If the reference does not end in an int, returns
        an empty list.
        """
        result = list()     # type: List[int]
        i = len(reference) - 1
        while isinstance(reference[i], int):
            result.insert(0, cast(int, reference[i]))
            i -= 1
        return result

    @staticmethod
    def __stem(reference: Reference) -> Reference:
        """Extracts the part of the reference before the slot.

        If there is no slot, returns the whole reference.
        """
        i = len(reference)
        while isinstance(reference[i-1], int):
            i -= 1
        return reference[:i]    # type: ignore


class MulticastConduit:
    """Multicast conduits connect multiple input ports to a single output port.

    In yMMSL they are expressed as a mapping:

    .. code-block:: yaml

        sender.port:
        - receiver1.port
        - receiver2.port

    This class is only used in the parsing and storing of the yMMSL file.
    Once parsed and populated in :class:`Model`, a multicast is identified by
    two or more conduits with the same :attr:`Conduit.sender`.
    """

    def __init__(self, sender: str, receiver: List[str]) -> None:
        """Create a Multicast Conduit.

        Args:
            sender: The sending component and port, as a Reference.
            receiver: The receiving components and ports, as a list of
                    References.
        """
        self.sender = sender
        # note: attribute must be called receiver to transparently work with
        # seq_attribute_to_map and map_attribute_to_seq in
        # Model._yatiml_savorize and Model._yatiml_sweeten
        self.receiver = receiver
        self._conduits = [Conduit(sender, recv) for recv in receiver]

    def as_conduits(self) -> List[Conduit]:
        """Retrieve the conduits that are part of this multicast conduit.

        Returns:
            A list of conduits, one conduit for each receiver.
        """
        return self._conduits


AnyConduit = Union[Conduit, MulticastConduit]


class ModelReference:
    """Describes a reference (by name) to a model.

    Attributes:
        name: The name of the simulation model this refers to.

    """
    def __init__(self, name: str) -> None:
        """Create a ModelReference.

        Arguments:
            name: Name of the model to refer to.

        """
        self.name = Identifier(name)


class Model(ModelReference):
    """Describes a simulation model.

    A model consists of a number of components connected by
    conduits.

    Note that there may be no conduits, if there is only a single
    component. In that case, the conduits argument may be
    omitted when constructing the object, and also from the YAML file;
    the `conduits` attribute will then be set to an empty list.

    Attributes:
        name: The name by which this simulation model is known to
                the system.
        components: A list of components making up the
                model.
        conduits: A list of conduits connecting the components.

    """
    def __init__(self, name: str,
                 components: List[Component],
                 conduits: Optional[List[AnyConduit]] = None) -> None:
        """Create a Model.

        Args:
            name: Name of this model.
            components: A list of components making up the model.
            conduits: A list of conduits connecting the components.
        """
        super().__init__(name)
        self.components = components

        self.conduits = list()      # type: List[Conduit]
        if conduits:
            for conduit in conduits:
                if isinstance(conduit, Conduit):
                    self.conduits.append(conduit)
                if isinstance(conduit, MulticastConduit):
                    self.conduits.extend(conduit.as_conduits())

    def update(self, overlay: 'Model') -> None:
        """Overlay another model definition on top of this one.

        This updates the object with the name, components and conduits
        given in the argument. The name is overwritten, and components
        are overwritten if they have the same name as an existing
        argument or else added.

        Conduits are added. If a receiving port was already connected, the
        old conduit is removed. If a sending port was already connected, the
        new conduit is added and the sending port acts as a multicast port.

        Args:
            overlay: A Model definition to overlay on top of this one.
        """
        self.name = overlay.name
        # update components
        for newc in overlay.components:
            for i, oldc in enumerate(self.components):
                if oldc.name == newc.name:
                    self.components[i] = newc
                    break
            else:
                self.components.append(newc)

        # remove overwritten conduits
        for newt in overlay.conduits:
            for oldt in self.conduits.copy():
                # Multiple conduits can be connected to one sending port
                # (multicast), only overwrite connections to a receiving port
                if oldt.receiver == newt.receiver:
                    self.conduits.remove(oldt)

        # add new conduits
        self.conduits.extend(overlay.conduits)

    def check_consistent(self) -> None:
        """Checks that the model is internally consistent.

        This checks whether all conduits are connected to existing
        components, and will raise a RuntimeError with an explanation
        if one is not.
        """
        def component_exists(name: Reference) -> bool:
            for comp in self.components:
                if comp.name == name:
                    return True
            return False

        def component_has_receiving_port(
                component: Reference, port: Identifier) -> bool:
            if port == 'muscle_settings_in':
                return True
            for comp in self.components:
                if comp.name == component:
                    if not comp.ports:
                        return True

                    try:
                        if comp.ports.operator(port).allows_receiving():
                            return True
                    except KeyError:
                        pass
            return False

        def component_has_sending_port(
                component: Reference, port: Identifier) -> bool:
            for comp in self.components:
                if comp.name == component:
                    if not comp.ports:
                        return True

                    try:
                        if comp.ports.operator(port).allows_sending():
                            return True
                    except KeyError:
                        pass
            return False

        receivers_seen = set()
        for conduit in self.conduits:
            scomp = conduit.sending_component()
            if not component_exists(scomp):
                raise RuntimeError(
                    'Unknown sending component "{}" of {}'.format(
                        scomp, conduit))

            rcomp = conduit.receiving_component()
            if not component_exists(rcomp):
                raise RuntimeError(
                    'Unknown receiving component "{}" of {}'.format(
                        rcomp, conduit))

            sport = conduit.sending_port()
            if not component_has_sending_port(scomp, sport):
                raise RuntimeError(
                        'Invalid conduit "{}": component "{}" does not'
                        ' have a sending port "{}"'.format(
                            conduit, scomp, sport))

            rport = conduit.receiving_port()
            if not component_has_receiving_port(rcomp, rport):
                raise RuntimeError(
                        'Invalid conduit "{}": component "{}" does not'
                        ' have a receiving port "{}"'.format(
                            conduit, rcomp, rport))

            if conduit.receiver in receivers_seen:
                raise RuntimeError(
                        'Receiving port "{}" is connected by multiple'
                        ' conduits.'.format(conduit.receiver))
            receivers_seen.add(conduit.receiver)

    def __conduits_for_export(self) -> List[AnyConduit]:
        """Process conduits and identify MulticastConduits for exporting.

        Returns:
            A list of Conduits and MulticastConduits.
        """
        cond_dct = OrderedDict()  # type: OrderedDict[Reference, List[Conduit]]
        for conduit in self.conduits:
            cond_dct.setdefault(conduit.sender, []).append(conduit)
        conduit_list = []   # type: List[AnyConduit]
        for sender, conduits in cond_dct.items():
            if len(conduits) == 1:
                conduit_list.append(conduits[0])
            else:
                conduit_list.append(MulticastConduit(
                        str(sender),
                        [str(conduit.receiver) for conduit in conduits]))
        return conduit_list

    def _yatiml_attributes(self) -> OrderedDict:
        return OrderedDict([
            ('name', self.name),
            ('components', self.components),
            ('conduits', self.__conduits_for_export())])

    @classmethod
    def _yatiml_recognize(cls, node: yatiml.UnknownNode) -> None:
        node.require_mapping()
        node.require_attribute('name')
        node.require_attribute('components')

    @classmethod
    def _yatiml_savorize(cls, node: yatiml.Node) -> None:
        node.map_attribute_to_seq('components', 'name', 'implementation')
        node.map_attribute_to_seq('conduits', 'sender', 'receiver')

    @classmethod
    def _yatiml_sweeten(cls, node: yatiml.Node) -> None:
        node.seq_attribute_to_map('components', 'name', 'implementation')
        if len(node.get_attribute('conduits').seq_items()) == 0:
            node.remove_attribute('conduits')
        node.seq_attribute_to_map('conduits', 'sender', 'receiver')



"""Definitions for specifying how to start a component."""
from enum import Enum
from pathlib import Path
from typing import cast, Dict, List, Optional, Union

from ruamel import yaml
import yatiml

from ymmsl.identity import Reference


class KeepsStateForNextUse(Enum):
    """Describes whether an implementation keeps internal state between
    iterations of the reuse loop.

    See also :ref:`Keeps state for next use`.
    """

    NECESSARY = 1
    """The implementation has an internal state that is necessary for
    continuing the implementation."""
    NO = 2
    """The implementation has no internal state."""
    HELPFUL = 3
    """The implementation has an internal state, though this could be
    regenerated. Doing so may be expensive."""

    @classmethod
    def _yatiml_savorize(cls, node: yatiml.Node) -> None:
        if node.is_scalar(str):
            val = cast(str, node.get_value())
            node.set_value(val.upper())

    @classmethod
    def _yatiml_sweeten(cls, node: yatiml.Node) -> None:
        val = node.get_value()
        if isinstance(val, str):
            node.set_value(val.lower())


class ExecutionModel(Enum):
    """Describes how to start a model component."""
    DIRECT = 1
    """Start directly on the allocated core(s), without MPI."""
    OPENMPI = 2
    """Start using OpenMPI's mpirun."""
    INTELMPI = 3
    """Start using Intel MPI's mpirun."""
    SRUNMPI = 4
    """Start MPI implementation using srun."""

    @classmethod
    def _yatiml_savorize(cls, node: yatiml.Node) -> None:
        if node.is_scalar(str):
            val = cast(str, node.get_value())
            node.set_value(val.upper())

    @classmethod
    def _yatiml_sweeten(cls, node: yatiml.Node) -> None:
        val = node.get_value()
        if isinstance(val, str):
            node.set_value(val.lower())


class Implementation:
    """Describes an installed implementation.

    An Implementation normally has an ``executable`` and any other
    needed attributes, with ``script`` set to None. You should specify
    a script only as a last resort, probably after getting some help
    from the authors of this library. If a script is specified, all
    other attributes except for the name must be ``None``.

    For ``execution_model``, the following values are supported:

    direct
      The program will be executed directly. Use this for non-MPI
      programs.

    openmpi
      For MPI programs that should be started using OpenMPI's mpirun.

    intelmpi
      For MPI programs that should be started using Intel MPI's
      mpirun.

    The ``can_share_resources`` attribute describes whether this
    implementation can share resources (cores) with other components
    in a macro-micro coupling. Set this to ``False`` if the
    implementation does significant computing inside of its time
    update loop after having sent messages on its O_I port(s) but
    before receiving messages on its S port(s). In the unlikely case
    that it's doing significant computing before receiving for F_INIT
    or after sending its O_F messages, likewise set this to ``False``.

    Setting this to ``False`` unnecessarily will waste core hours,
    setting it to ``True`` incorrectly will slow down your simulation.

    Attributes:
        name: Name of the implementation
        modules: HPC software modules to load
        virtual_env: Path to a virtual env to activate
        env: Environment variables to set
        execution_model: How to start the executable
        executable: Full path to executable to run
        args: Arguments to pass to the executable
        script: A script that starts the implementation
        can_share_resources: Whether this implementation can share
            resources (cores) with other components or not
        keeps_state_for_next_use: Does this implementation keep state for the
            next iteration of the reuse loop. See :class:`ImplementationState`.
    """

    def __init__(
            self,
            name: Reference,
            modules: Union[str, List[str], None] = None,
            virtual_env: Optional[Path] = None,
            env: Optional[Dict[str, str]] = None,
            execution_model: ExecutionModel = ExecutionModel.DIRECT,
            executable: Optional[Path] = None,
            args: Union[str, List[str], None] = None,
            script: Union[str, List[str], None] = None,
            can_share_resources: bool = True,
            keeps_state_for_next_use: KeepsStateForNextUse
            = KeepsStateForNextUse.NECESSARY
            ) -> None:
        """Create an Implementation description.

        An Implementation normally has an ``executable`` and any other
        needed arguments, with ``script`` set to ``None``. You should
        specify a script only as a last resort, probably after getting
        some help from the authors of this library. If ``script`` is
        specified, all other arguments except for ``name`` must be
        ``None``.

        If script is a list, each string in it is a line, and the
        lines will be concatenated into a single string to put into
        the script attribute.

        Args:
            name: Name of the implementation
            modules: HPC software modules to load
            virtual_env: Path to a virtual env to activate
            env: Environment variables to set
            execution_model: How to start the executable, see above.
            executable: Full path to executable to run
            args: Arguments to pass to the executable
            script: Script that starts the implementation
            can_share_resources: Whether this implementation can share
                    resources (cores) with other components or not.
                    See above.
            keeps_state_for_next_use: Does this implementation keep state for
                the next iteration of the reuse loop. See
                :class:`ImplementationState`.
        """
        if script is not None:
            if (
                    modules is not None or virtual_env is not None or
                    env is not None or
                    execution_model is not ExecutionModel.DIRECT or
                    executable is not None or args is not None):
                raise RuntimeError(
                        'When creating an Implementation, script was specified'
                        ' together with another argument, which is not'
                        ' supported. Please specify either a script or an'
                        ' executable.')

        if executable is not None and script is not None:
            raise RuntimeError(
                    f'In {name}, both a script and an executable were given.'
                    ' Please specify either a script, or the other parameters.'
                    )

        if executable is None and script is None:
            raise RuntimeError(
                    f'In {name}, neither a script nor an executable was given.'
                    ' Please specify either a script, or the other parameters.'
                    )

        self.name = name

        if isinstance(script, list):
            self.script = '\n'.join(script) + '\n'  # type: Optional[str]
        else:
            self.script = script

        if isinstance(modules, str):
            self.modules = modules.split(' ')   # type: Optional[List[str]]
        else:
            self.modules = modules
        self.virtual_env = virtual_env
        if env is None:
            env = dict()
        self.env = env
        self.execution_model = execution_model
        self.executable = executable

        if isinstance(args, str):
            self.args = [args]  # type: Optional[List[str]]
        else:
            self.args = args

        self.can_share_resources = can_share_resources
        self.keeps_state_for_next_use = keeps_state_for_next_use

    @classmethod
    def _yatiml_recognize(cls, node: yatiml.UnknownNode) -> None:
        # There's no ambiguity, and we want to allow some leeway
        # and savorize things, so disable recognition.
        pass

    @classmethod
    def _yatiml_savorize(cls, node: yatiml.Node) -> None:
        if node.has_attribute('env'):
            env_node = node.get_attribute('env')
            if env_node.is_mapping():
                for _, value_node in env_node.yaml_node.value:
                    if isinstance(value_node, yaml.ScalarNode):
                        if value_node.tag == 'tag:yaml.org,2002:int':
                            value_node.tag = 'tag:yaml.org,2002:str'
                        if value_node.tag == 'tag:yaml.org,2002:float':
                            value_node.tag = 'tag:yaml.org,2002:str'
                        if value_node.tag == 'tag:yaml.org,2002:bool':
                            value_node.tag = 'tag:yaml.org,2002:str'

    _yatiml_defaults = {
        'execution_model': 'direct',
        'keeps_state_for_next_use': 'necessary'}

    @classmethod
    def _yatiml_sweeten(cls, node: yatiml.Node) -> None:
        if node.has_attribute('script'):
            script_node = node.get_attribute('script')
            if script_node.is_scalar(str):
                text = cast(str, script_node.get_value())
                if '\n' in text:
                    cast(yaml.ScalarNode, script_node.yaml_node).style = '|'

        node.remove_attributes_with_default_values(cls)
        if node.has_attribute('env'):
            env_attr = node.get_attribute('env')
            if env_attr.is_mapping():
                if env_attr.is_empty():
                    node.remove_attribute('env')


class ResourceRequirements:
    """Describes resources to allocate for components.

    Attributes:
        name: Name of the component to configure.
    """
    def __init__(self, name: Reference) -> None:
        """Create a ResourceRequirements description.

        Args:
            name: Name of the component to configure.
        """
        self.name = name

    @classmethod
    def _yatiml_recognize(cls, node: yatiml.UnknownNode) -> None:
        raise yatiml.RecognitionError(
                'Please specify either "threads" or "mpi_processes".')


class ThreadedResReq(ResourceRequirements):
    """Describes resources for threaded implementations.

    This includes singlethreaded and multithreaded implementations
    that do not support MPI. As many cores as specified will be
    allocated on a single node, for each instance.

    Attributes:
        name: Name of the component to configure.
        threads: Number of threads/cores per instance.
    """

    def __init__(self, name: Reference, threads: int) -> None:
        """Create a ThreadedResourceRequirements description.

        Args:
            name: Name of the component to configure.
            threads: Number of threads (cores) per instance.
        """
        super().__init__(name)
        self.threads = threads


class MPICoresResReq(ResourceRequirements):
    """Describes resources for simple MPI implementations.

    This allocates individual cores or sets of cores on the same node
    for a given number of MPI processes per instance.

    Attributes:
        name: Name of the component to configure.
        mpi_processes: Number of MPI processes to start.
        threads_per_mpi_process: Number of threads/cores per process.
    """

    def __init__(
            self, name: Reference, mpi_processes: int,
            threads_per_mpi_process: int = 1) -> None:
        """Create a ThreadedMPIResourceRequirements description.

        Args:
            name: Name of the component to configure.
            mpi_processes: Number of MPI processes to start.
            threads_per_mpi_process: Number of threads/cores per
                    process. Defaults to 1.
        """
        super().__init__(name)
        self.mpi_processes = mpi_processes
        self.threads_per_mpi_process = threads_per_mpi_process

    @classmethod
    def _yatiml_sweeten(cls, node: yatiml.Node) -> None:
        node.remove_attributes_with_default_values(cls)


class MPINodesResReq(ResourceRequirements):
    """Describes resources for node based MPI implementations.

    This allocates resources for an MPI process in terms of nodes and
    cores, processes and threads on them.

    Attributes:
        name: Name of the component to configure.
        nodes: Number of nodes to reserve.
        mpi_processes_per_node: Number of MPI processes to start on
                each node.
        threads_per_mpi_process: Number of threads/cores per process.
    """

    def __init__(
            self, name: Reference, nodes: int,
            mpi_processes_per_node: int, threads_per_mpi_process: int = 1
            ) -> None:
        """Create a NodeBasedMPIResourceRequirements description.

        Args:
            name: Name of the component to configure.
            nodes: Number of nodes to reserve.
            mpi_processes_per_node: Number of MPI processes to start
                    on each node.
            threads_per_mpi_process: Number of threads/cores per
                    process. Defaults to 1.
        """
        super().__init__(name)
        self.nodes = nodes
        self.mpi_processes_per_node = mpi_processes_per_node
        self.threads_per_mpi_process = threads_per_mpi_process

    @classmethod
    def _yatiml_sweeten(cls, node: yatiml.Node) -> None:
        node.remove_attributes_with_default_values(cls)



"""Definitions for describing checkpoints."""

from typing import List, Optional, Union

from ruamel import yaml
import yatiml


class CheckpointRule:
    """Defines a checkpoint rule.

    There are two flavors of rules: :class:`CheckpointRangeRule` and
    :class:`CheckpointAtRule`. Do not use this class directly.
    """


class CheckpointRangeRule(CheckpointRule):
    """Defines a range of checkpoint moments.

    If ``start`` is supplied, this rule specifies a checkpoint at: ``start``,
    ``start + every``, ``start + 2*every``, etc., for as long as
    ``start + n*every <= stop``, with ``n`` a whole number. If stop is not
    given, the range continues indefinitely.

    Start may be omitted, in which case a checkpoint is defined for ``0``,
    ``every``, ``2*every``, etc. Note that in this case the range also extends
    to negative numbers (``-every``, ``-2*every``, etc.), as simulated time may
    be negative (e.g. in rocket launch, ``t=0`` is generally taken as lift-off
    time, but events already take place before that moment).

    Attributes:
        start: Start of the range.
        stop: Stopping criterium of the range.
        every: Step size of the range, must be positive.
    """

    def __init__(self,
                 start: Optional[Union[float, int]] = None,
                 stop: Optional[Union[float, int]] = None,
                 every: Union[float, int] = 0) -> None:
        """Create a checkpoint range.

        Args:
            start: Start of the range. Defaults to None.
            stop: Stop criterium of the range. Defaults to None.
            every: Step size, must be larger than 0.
        """
        if every <= 0:
            raise ValueError(f"Invalid every {every} in checkpoint range:"
                             " must be larger than 0.")
        if start is not None and stop is not None and start > stop:
            raise ValueError(f"Invalid start {start} and stop {stop} in"
                             "checkpoint range: stop cannot be smaller than"
                             " start.")
        self.start = start
        self.stop = stop
        self.every = every

    @classmethod
    def _yatiml_recognize(cls, node: yatiml.UnknownNode) -> None:
        node.require_attribute('every')

    @classmethod
    def _yatiml_sweeten(cls, node: yatiml.Node) -> None:
        if node.get_attribute("start").is_scalar(type(None)):
            node.remove_attribute("start")

        if node.get_attribute("stop").is_scalar(type(None)):
            node.remove_attribute("stop")


class CheckpointAtRule(CheckpointRule):
    """Defines an "at" checkpoint rule.

    An "at" checkpoint rule creates a snapshot at the specified moments.

    Attributes:
        at: List of checkpoints.
    """

    def __init__(self, at: Optional[List[Union[float, int]]]) -> None:
        """Create checkpoint rules.

        Args:
            at: List of checkpoints. Defaults to None.
        """
        if at is None:
            at = []
        self.at = at
        self.at.sort()

    @classmethod
    def _yatiml_recognize(cls, node: yatiml.UnknownNode) -> None:
        node.require_attribute('at')

    @classmethod
    def _yatiml_savorize(cls, node: yatiml.Node) -> None:
        if node.has_attribute('at'):
            if (node.has_attribute_type('at', int) or
                    node.has_attribute_type('at', float)):
                attr = node.get_attribute('at')
                start_mark = attr.yaml_node.start_mark
                end_mark = attr.yaml_node.end_mark
                new_seq = yaml.nodes.SequenceNode(
                        'tag:yaml.org,2002:seq', [attr.yaml_node], start_mark,
                        end_mark)
                node.set_attribute('at', new_seq)


class Checkpoints:
    """Defines checkpoints in a configuration.

    There are three checkpoint triggers: `at_end`, `wallclock_time` and
    `simulation_time`. The `at_end` trigger specifies that a checkpoint should
    be created just before the workflow finishes. The `wallclock_time` trigger
    is based on the elapsed real time since starting the muscle_manager in
    seconds. The `simulation_time` trigger is based on the time in the
    simulation as reported by the instances.

    Note that the `simulation_time` trigger assumes a shared concept of time
    among all components of the model.

    Attributes:
        at_end: Whether a checkpoint should be created just before ending the
            workflow.
        wallclock_time: Checkpoint rules for the wallclock_time trigger.
        simulation_time: Checkpoint rules for the simulation_time trigger.
    """
    def __init__(self,
                 at_end: bool = False,
                 wallclock_time: Optional[List[CheckpointRule]] = None,
                 simulation_time: Optional[List[CheckpointRule]] = None
                 ) -> None:
        """Create checkpoint definitions.

        Args:
            wallclock_time: Checkpoint rules for the wallclock_time trigger.
            simulation_time: Checkpoint rules for the simulation_time trigger.
        """
        self.at_end = at_end
        if wallclock_time is None:
            wallclock_time = []
        if simulation_time is None:
            simulation_time = []
        self.wallclock_time = wallclock_time
        self.simulation_time = simulation_time

    def __bool__(self) -> bool:
        """Evaluate to true iff any rules are defined."""
        return (self.at_end or
                bool(self.wallclock_time) or
                bool(self.simulation_time))

    def update(self, overlay: 'Checkpoints') -> None:
        """Update this checkpoints with the given overlay.

        Sets `at_end` to True if it is set in the overlay, otherwise `at_end`
        remains as is. Updates the checkpoint rules for wallclock time and
        simulation time. See :meth:`CheckpointRules.update`.
        """
        self.at_end |= overlay.at_end
        self.wallclock_time.extend(overlay.wallclock_time)
        self.simulation_time.extend(overlay.simulation_time)

    @classmethod
    def _yatiml_sweeten(cls, node: yatiml.Node) -> None:
        if not node.get_attribute("at_end").get_value():
            node.remove_attribute("at_end")

        wallclock_time = node.get_attribute("wallclock_time")
        if wallclock_time.is_scalar(type(None)) or (
                wallclock_time.is_sequence() and wallclock_time.is_empty()):
            node.remove_attribute("wallclock_time")

        simulation_time = node.get_attribute("simulation_time")
        if simulation_time.is_scalar(type(None)) or (
                simulation_time.is_sequence() and simulation_time.is_empty()):
            node.remove_attribute("simulation_time")



"""This module contains definitions for identity."""
from copy import copy
import re
from collections import UserString
from typing import Any, Generator, Iterable, List, overload, Union

import yatiml


class Identifier(UserString):
    """A custom string type that represents an identifier.

    An identifier may consist of upper- and lowercase characters, digits, and \
    underscores.
    """

    def __init__(self, seq: Any) -> None:
        """Create an Identifier.

        This creates a new identifier object, using the string
        representation of whichever object you pass.

        Raises:
            ValueError: If the argument's string representation does
                    not form a valid Identifier.

        """
        super().__init__(seq)
        if not re.fullmatch(
                r'[a-zA-Z_]\w*', self.data, flags=re.ASCII):    # type: ignore
            raise ValueError('Identifiers must consist only of'
                             ' lower- and uppercase letters, digits and'
                             ' underscores, must start with a letter or'
                             ' an underscore, and must not be empty.')


ReferencePart = Union[Identifier, int]


class Reference(yatiml.String):
    """A reference to an object in the MMSL execution model.

    References in string form are written as either:

    -  an Identifier,
    -  a Reference followed by a period and an Identifier, or
    -  a Reference followed by an integer enclosed in square brackets.

    In object form, they consist of a list of Identifiers and ints. The \
    first list item is always an Identifier. For the rest of the list, \
    an Identifier represents a period operator with that argument, \
    while an int represents the indexing operator with that argument.

    Reference objects act like a list of Identifiers and ints, you can
    get their length using len(), iterate through the parts using a
    loop, and get sublists or individual items using []. Note that the
    sublist has to be a valid Reference, so it cannot start with an
    int.

    References can be compared for equality to each other or to a
    plain string, and they can be used as dictionary keys. Reference
    objects are immutable (or they're supposed to be anyway), so do not
    try to change any of the elements. Instead, make a new Reference.
    Especially References that are used as dictionary keys must not be
    modified, this will get your dictionary in a very confused state.
    """

    def __init__(self, parts: Union[str, List[ReferencePart]]) -> None:
        """Create a Reference.

        Creates a Reference from either a string, which will be parsed,
        or a list of Identifiers and ints.

        Args:
            parts: Either a list of parts, or a string to parse.

        Raises:
            ValueError: If the argument does not define a valid
                    Reference.

        """
        if isinstance(parts, str):
            self._parts = self._string_to_parts(parts)
        elif len(parts) > 0 and not isinstance(parts[0], Identifier):
            raise ValueError(
                    'The first part of a Reference must be an Identifier')
        else:
            self._parts = parts

    def __str__(self) -> str:
        """Convert the Reference to string form."""
        return self._parts_to_string(self._parts)

    def __repr__(self) -> str:
        """Produce a representation in string form."""
        return 'Reference("{}")'.format(str(self))

    def __len__(self) -> int:
        """Return the number of parts in the Reference."""
        return len(self._parts)

    def __hash__(self) -> int:
        """Calculate a hash value for use by dicts."""
        return hash(str(self))

    def __eq__(self, other: Any) -> bool:
        """Compare for equality.

        Will compare part-by-part if the other argument is a Reference,
        or string representations if the other argument is a string.

        Args:
            other: Another Reference or a string.

        """
        if isinstance(other, Reference):
            return self._parts == other._parts
        if isinstance(other, str):
            return str(self) == other
        return NotImplemented

    def __ne__(self, other: Any) -> bool:
        """Compare for equality.

        Will compare part-by-part if the other argument is a Reference,
        or string representations if the other argument is a string.

        Args:
            other: Another Reference or a string.

        """
        if isinstance(other, Reference):
            return self._parts != other._parts
        if isinstance(other, str):
            return str(self) != other
        return NotImplemented

    def __lt__(self, other: Any) -> bool:
        """Order alphabetically.

        Args:
            other: Another Reference or a string.
        """
        if isinstance(other, Reference):
            i = 0
            while i < min(len(self._parts), len(other._parts)):
                s = self._parts[i]
                o = other._parts[i]

                if isinstance(s, int) and isinstance(o, int):
                    if s < o:
                        return True
                    if o < s:
                        return False
                elif isinstance(s, Identifier) and isinstance(o, Identifier):
                    if s < o:       # repeated because mypy is dumb
                        return True
                    if o < s:
                        return False
                elif isinstance(s, int) and isinstance(o, Identifier):
                    return True
                elif isinstance(s, Identifier) and isinstance(o, int):
                    return False
                i += 1
            return len(self._parts) < len(other._parts)

        if isinstance(other, str):
            return self < Reference(other)

        return NotImplemented

    def __iter__(self) -> Generator[ReferencePart, None, None]:
        """Iterate through the parts.

        Yields:
            Each part in turn from left to right.

        """
        for part in self._parts:
            yield part

    @overload
    def __getitem__(self, key: int) -> ReferencePart: ...

    @overload
    def __getitem__(self, key: slice) -> 'Reference': ...

    def __getitem__(
            self, key: Union[int, slice]
            ) -> Union['Reference', ReferencePart]:
        """Get a part or a slice.

        If passed an int, e.g. ref[2], will return that part as an int
        or an Identifier. If passed a slice, e.g. ref[2:], will return
        a Reference comprising those parts. This must be a valid
        Reference, so it must start with an Identifier.

        Args:
            key: Which part to return.

        Returns:
            A part or a sub-Reference.

        Raises:
            ValueError: If the argument is not an int or a slice.

        """
        if isinstance(key, int):
            return self._parts[key]
        if isinstance(key, slice):
            return Reference(self._parts[key])
        raise ValueError('Subscript must be either an int or a slice')

    def __setitem__(self, key: Union[int, slice], value: Any) -> None:
        """Does not set the value of a part.

        References are immutable, so they should not be modified, and
        this method just gives an error.

        Raises:
            RuntimeError: Always.

        """
        raise RuntimeError('Reference objects are immutable, please don\'t try'
                           ' to change them.')

    def __add__(self, other: Union['Reference', Iterable[ReferencePart],
                ReferencePart]) -> 'Reference':
        """Concatenates something onto a Reference.

        The object to add on can be another Reference, a list of
        reference parts (either int or Identifier), or an int or an
        Identifier. A new Reference will be created equal to the
        original one with the new part attached on the right.

        Args:
            other: The object to concatenate.

        Returns:
            A new concatenated Reference.

        """
        ret = Reference(copy(self._parts))
        if isinstance(other, Reference):
            ret._parts.extend(other._parts)
        elif isinstance(other, (Identifier, int)):
            ret._parts.append(other)
        elif hasattr(other, '__iter__'):
            ret._parts.extend(other)
        return ret

    def without_trailing_ints(self) -> 'Reference':
        """Returns a copy of this Reference with trailing ints removed.

        Examples:
            a.b.c[1][2] -> a.b.c
            a[1].b.c -> a[1].b.c
            a.b.c -> a.b.c
            a[1].b.c[2] -> a[1].b.c
        """
        i = len(self._parts) - 1
        while i > 0 and isinstance(self._parts[i], int):
            i -= 1
        return Reference(self._parts[0:i+1])

    @classmethod
    def _string_to_parts(cls, text: str) -> List[ReferencePart]:
        """Parse a string into a list of parts.

        Args:
            text: The string to parse.

        Raises:
            ValueError: If the string does not represent a valid
                    Reference.

        """
        def find_next_op(text: str, start: int) -> int:
            next_bracket = text.find('[', start)
            if next_bracket == -1:
                next_bracket = len(text)
            next_period = text.find('.', start)
            if next_period == -1:
                next_period = len(text)
            return min(next_period, next_bracket)

        end = len(text)
        cur_op = find_next_op(text, 0)
        parts = [Identifier(text[0:cur_op])]  # type: List[ReferencePart]
        while cur_op < end:
            if text[cur_op] == '.':
                next_op = find_next_op(text, cur_op + 1)
                parts.append(Identifier(text[cur_op + 1:next_op]))
                cur_op = next_op
            elif text[cur_op] == '[':
                close_bracket = text.find(']', cur_op)
                if close_bracket == -1:
                    raise ValueError('Missing closing bracket in Reference {}'
                                     ''.format(text))
                try:
                    index = int(text[cur_op + 1:close_bracket])
                except ValueError as exc:
                    raise ValueError('Invalid index \'{}\' in {}, expected an'
                                     ' int'.format(
                                         text[cur_op + 1:close_bracket], text)
                                     ) from exc
                parts.append(index)
                cur_op = close_bracket + 1
            else:
                raise ValueError('Invalid character \'{}\' encountered in'
                                 ' Reference {}'.format(text[cur_op], text))
        return parts

    @classmethod
    def _parts_to_string(cls, parts: List[ReferencePart]) -> str:
        """Convert a list of parts to its string representation.

        Args:
            parts: The parts to represent.

        Returns:
            Their string form.

        """
        text = str(parts[0])
        for part in parts[1:]:
            if isinstance(part, int):
                text += '[{}]'.format(part)
            else:
                text += '.{}'.format(part)
        return text



"""This module contains all the definitions for yMMSL."""
from collections import OrderedDict
import collections.abc as abc
import logging
from pathlib import Path
from typing import (
        Dict, List, MutableMapping, Optional, Sequence, Union, cast)

import yatiml
from ruamel import yaml

from ymmsl.checkpoint import Checkpoints
from ymmsl.document import Document
from ymmsl.identity import Reference
from ymmsl.execution import (
        ExecutionModel, Implementation,
        ResourceRequirements, ThreadedResReq)
from ymmsl.settings import Settings
from ymmsl.model import Model, ModelReference


_logger = logging.getLogger(__name__)

# This should be a local definition, but that triggers a mypy issue:
# https://github.com/python/mypy/issues/7281
_ImplType = Dict[Reference, Implementation]


_ResType = MutableMapping[Reference, ResourceRequirements]


class PartialConfiguration(Document):
    """Top-level class for all information in a yMMSL file.

    Attributes:
        model: A model to run.
        settings: Settings to run the model with.
        implementations: Implementations to use to run the model.
            Dictionary mapping implementation names (as References) to
            Implementation objects.
        resources: Resources to allocate for the model components.
            Dictionary mapping component names to ResourceRequirements
            objects.
        description: A human-readable description of the configuration.
        checkpoints: Defines when each model component should create a snapshot
        resume: Defines what snapshot each model component should resume from
    """
    def __init__(self,
                 model: Optional[ModelReference] = None,
                 settings: Optional[Settings] = None,
                 implementations: Optional[Union[
                     List[Implementation],
                     Dict[Reference, Implementation]]] = None,
                 resources: Optional[Union[
                     Sequence[ResourceRequirements],
                     MutableMapping[Reference, ResourceRequirements]]] = None,
                 description: Optional[str] = None,
                 checkpoints: Optional[Checkpoints] = None,
                 resume: Optional[Dict[Reference, Path]] = None
                 ) -> None:
        """Create a Configuration.

        Implementations and resources may be either a list of such
        objects, or a dictionary matching the attribute format (see
        above).

        Args:
            model: A description of the model to run.
            settings: Settings to run the model with.
            implementations: Implementations to choose from.
            resources: Resources to allocate for the model components.
            description: Human-readable description
            checkpoints: When each component should create a snapshot
            resume: What snapshot each component should resume from

        """
        self.model = model

        if settings is None:
            self.settings = Settings()
        else:
            self.settings = settings

        if implementations is None:
            self.implementations = dict()   # type: _ImplType
        elif isinstance(implementations, list):
            self.implementations = OrderedDict([
                (impl.name, impl) for impl in implementations])
        else:
            self.implementations = implementations

        if resources is None:
            self.resources = dict()     # type: _ResType
        elif isinstance(resources, abc.Sequence):
            self.resources = OrderedDict([
                (res.name, res) for res in resources])
        else:
            self.resources = resources

        if description is None:
            self.description = ''
        else:
            self.description = description

        if checkpoints is None:
            self.checkpoints = Checkpoints()
        else:
            self.checkpoints = checkpoints

        if resume is None:
            self.resume = dict()    # type: Dict[Reference, Path]
        else:
            self.resume = resume

    def update(self, overlay: 'PartialConfiguration') -> None:
        """Update this configuration with the given overlay.

        This will update the model according to :meth:`Model.update`, copy
        settings from overlay on top of the current settings, overwrite
        implementations with the same name and add implementations with a new
        name, and likewise for resources and resume. The description of the
        overlay is appended to the current description. Checkpoints are updated
        according to :meth:`Checkpoints.update`.

        Args:
            overlay: A configuration to overlay onto this one.
        """
        # beware of isinstance matching parent classes...
        if self.model is None:
            self.model = overlay.model
        elif isinstance(overlay.model, Model):
            if isinstance(self.model, Model):
                self.model.update(overlay.model)
            else:
                # self.model is a ModelReference
                self.model = overlay.model
        elif isinstance(overlay.model, ModelReference):
            self.model.name = overlay.model.name

        self.settings.update(overlay.settings)
        self.implementations.update(overlay.implementations)
        self.resources.update(overlay.resources)

        if not self.description:
            self.description = overlay.description
        elif overlay.description:
            self.description += '\n\n' + overlay.description

        self.checkpoints.update(overlay.checkpoints)
        self.resume.update(overlay.resume)

    def as_configuration(self) -> 'Configuration':
        """Converts to a full Configuration object.

        This checks that this PartialConfiguration has all the pieces
        needed to run a simulation, and if so converts it to a
        Configuration object.

        Note that this doesn't check references, just that there is
        a model, implementations and resources. For the more extensive
        check, see :meth:`Configuration.check_consistent`.

        Returns:
            A corresponding Configuration.

        Raises:
            ValueError: If this configuration isn't complete.
        """
        if (
                self.model is not None and isinstance(self.model, Model) and
                self.implementations and self.resources):
            return Configuration(
                    self.model, self.settings, self.implementations,
                    self.resources, self.description, self.checkpoints,
                    self.resume)
        raise ValueError(
                'Model, implementations or resources are missing from the'
                ' configuration.')

    @classmethod
    def _yatiml_recognize(cls, node: yatiml.UnknownNode) -> None:
        # Because of the syntactic sugar below, the YAML doesn't match
        # the types declared above. So we override recognition, and
        # since the attributes are all optional, we have nothing to
        # check except that it's a valid Document (with a version tag).
        Document._yatiml_recognize(node)

    @classmethod
    def _yatiml_savorize(cls, node: yatiml.Node) -> None:
        if not node.has_attribute('settings'):
            node.set_attribute('settings', None)
        node.map_attribute_to_index('implementations', 'name', 'script')
        node.map_attribute_to_index('resources', 'name')

    @classmethod
    def _yatiml_sweeten(cls, node: yatiml.Node) -> None:
        if node.get_attribute('model').is_scalar(type(None)):
            node.remove_attribute('model')

        if node.get_attribute('settings').is_scalar(type(None)):
            node.remove_attribute('settings')
        if len(node.get_attribute('settings').yaml_node.value) == 0:
            node.remove_attribute('settings')

        impl = node.get_attribute('implementations')
        if (
                impl.is_scalar(type(None)) or
                impl.is_mapping() and impl.is_empty()):
            node.remove_attribute('implementations')
        node.index_attribute_to_map('implementations', 'name', 'script')

        res = node.get_attribute('resources')
        if (
                res.is_scalar(type(None)) or
                res.is_mapping() and res.is_empty()):
            node.remove_attribute('resources')
        node.index_attribute_to_map('resources', 'name')

        descr = node.get_attribute('description')
        if descr.is_scalar(type(None)) or descr.get_value() == '':
            node.remove_attribute('description')
        if descr.is_scalar(str) and '\n' in cast(str, descr.get_value()):
            # output multi-line string in literal mode
            cast(yaml.ScalarNode, descr.yaml_node).style = '|'

        cp = node.get_attribute('checkpoints')
        if (
                cp.is_scalar(type(None)) or
                cp.is_mapping() and cp.is_empty()):
            node.remove_attribute('checkpoints')

        resu = node.get_attribute('resume')
        if (
                resu.is_scalar(type(None)) or
                resu.is_mapping() and resu.is_empty()):
            node.remove_attribute('resume')


class Configuration(PartialConfiguration):
    """Configuration that includes all information for a simulation.

    PartialConfiguration has some optional attributes, because we want
    to allow configuration files which only contain some of the
    information needed to run a simulation. At some point however,
    you need all the bits, and this class requires them.

    When loading a yMMSL file, you will automatically get an object
    of this class if all the required components are there; if the
    file is incomplete, you'll get a PartialConfiguration instead.

    Attributes:
        model: A model to run.
        settings: Settings to run the model with.
        implementations: Implementations to use to run the model.
            Dictionary mapping implementation names (as References) to
            Implementation objects.
        resources: Resources to allocate for the model components.
            Dictionary mapping component names to Resources objects.
        description: A human-readable description of the configuration.
        checkpoints: Defines when each model component should create a snapshot
        resume: Defines what snapshot each model component should resume from
    """
    def __init__(self,
                 model: Model,
                 settings: Optional[Settings] = None,
                 implementations: Union[
                     List[Implementation],
                     Dict[Reference, Implementation]] = [],
                 resources: Union[
                     Sequence[ResourceRequirements],
                     MutableMapping[Reference, ResourceRequirements]] = [],
                 description: Optional[str] = None,
                 checkpoints: Optional[Checkpoints] = None,
                 resume: Optional[Dict[Reference, Path]] = None
                 ) -> None:
        """Create a Configuration.

        Implementations and resources may be either a list of such
        objects, or a dictionary matching the attribute format (see
        above).

        Args:
            model: A description of the model to run.
            settings: Settings to run the model with.
            implementations: Implementations to choose from.
            resources: Resources to allocate for the model components.
            description: Human-readable description
            checkpoints: When each component should create a snapshot
            resume: What snapshot each component should resume from

        """
        # Make mypy recognize that, unlike PartialConfiguration.model,
        # Configuration.model always has type Model
        self.model: Model
        super().__init__(
                model, settings, implementations, resources, description,
                checkpoints, resume)

    def check_consistent(self) -> None:
        """Checks that the configuration is internally consistent.

        This checks whether all conduits are connected to existing
        components, that there's an implementation for every component,
        and that resources have been requested for each component.

        If any of these requirements is false, this function will
        raise a RuntimeError with an explanation of the problem.

        """
        self.model.check_consistent()

        for comp in self.model.components:
            if comp.implementation not in self.implementations:
                raise RuntimeError((
                        'Model component {} is missing an'
                        ' implementation').format(comp))
            if comp.name not in self.resources:
                raise RuntimeError((
                        'Model component {} is missing a resource'
                        ' allocation.').format(comp))

            impl = self.implementations[comp.implementation]
            res = self.resources[comp.name]
            if impl.execution_model == ExecutionModel.DIRECT:
                if not isinstance(res, ThreadedResReq):
                    raise RuntimeError((
                        'Model component {}\'s implementation does not'
                        ' specify MPI, but mpi_processes are specified in its'
                        ' resources. Please either set "execution_model" to'
                        ' an MPI model, or specify a number of threads.'
                        ).format(comp))
            else:
                if isinstance(res, ThreadedResReq):
                    raise RuntimeError((
                        'Model component {}\'s implementation specifies MPI,'
                        ' but threads are specified in its resources. Please'
                        ' either set "execution_model" to "direct", or'
                        ' specify a number of mpi processes.').format(comp))

    @classmethod
    def _yatiml_recognize(cls, node: yatiml.UnknownNode) -> None:
        node.require_attribute('model', Model)
        node.require_attribute('implementations')
        node.require_attribute('resources')



"""Loading and saving functions."""
from pathlib import Path
from typing import Any, IO, Union

import yatiml

from ymmsl.checkpoint import (
        CheckpointRule, CheckpointRangeRule, CheckpointAtRule, Checkpoints)
from ymmsl.component import Component, Ports
from ymmsl.configuration import Configuration, PartialConfiguration
from ymmsl.document import Document
from ymmsl.execution import (
        ExecutionModel, Implementation, ResourceRequirements, MPICoresResReq,
        MPINodesResReq, ThreadedResReq, KeepsStateForNextUse)
from ymmsl.settings import Settings
from ymmsl.identity import Identifier, Reference
from ymmsl.model import Conduit, MulticastConduit, Model, ModelReference


_classes = (
        PartialConfiguration, CheckpointRangeRule, CheckpointAtRule,
        CheckpointRule, Checkpoints, Component, Conduit, Configuration,
        Document, ExecutionModel, Identifier, Implementation,
        KeepsStateForNextUse, Model, ModelReference,
        MPICoresResReq, MPINodesResReq, Ports, Reference, ResourceRequirements,
        Settings, ThreadedResReq, MulticastConduit)


_load = yatiml.load_function(*_classes)


def load(source: Union[str, Path, IO[Any]]) -> PartialConfiguration:
    """Loads a yMMSL document from a string or a file.

    Args:
        source: A string containing yMMSL data, a pathlib Path to a
                file containing yMMSL data, or an open file-like
                object containing from which yMMSL data can be read.

    Returns:
        A PartialConfiguration object corresponding to the input data.

    """
    # This wrapper is just here to render the documentation.
    return _load(source)


_dump = yatiml.dumps_function(*_classes)


def dump(config: PartialConfiguration) -> str:
    """Converts a yMMSL configuration to a string containing YAML.

    Args:
        config: The configuration to be saved to yMMSL.

    Returns:
        A yMMSL YAML description of the given document.

    """
    # This wrapper is just here to render the documentation.
    return _dump(config)


_save = yatiml.dump_function(*_classes)


def save(
        config: PartialConfiguration, target: Union[str, Path, IO[Any]]
        ) -> None:
    """Saves a yMMSL configuration to a file.

    Args:
        config: The configuration to save to yMMSL.
        target: The file to save to, either as a string containing a
            path, as a pathlib Path object, or as an open file-like
            object.

    """
    # This wrapper is just here to render the documentation.
    _save(config, target)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

setup(
    name='ymmsl',
    version='0.13.0',
    description="Python bindings for the YAML version of the Multiscale Modeling and Simulation Language",
    long_description=readme + '\n\n',
    author="Lourens Veen",
    author_email='l.veen@esciencecenter.nl',
    url='https://github.com/multiscale/ymmsl-python',
    packages=[
        'ymmsl',
    ],
    package_dir={'ymmsl':
                 'ymmsl'},
    include_package_data=True,
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords='yMMSL multiscale modeling simulation YAML',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    python_requires='>=3.7, <4',
    test_suite='tests',
    install_requires=[
        'ruamel.yaml<=0.17.21,!=0.17.5',
        'yatiml>=0.10.0,<0.11.0'
    ],
)



"""Definitions for describing simulation components."""
from collections import OrderedDict
from enum import Enum
import logging
from typing import Dict     # noqa: F401
from typing import Iterable, List, Optional, Union

import ruamel.yaml as yaml
import yatiml

from ymmsl.identity import Identifier, Reference


_logger = logging.getLogger(__name__)


class Operator(Enum):
    """An operator of a component.

    This is a combination of the Submodel Execution Loop operators,
    and operators for other components such as mappers.
    """

    NONE = 0    #: No operator
    F_INIT = 1  #: Initialisation phase, before start of the SEL
    O_I = 2     #: State observation within the model's main loop
    S = 3       #: State update in the model's main loop
    O_F = 5     #: Observation of final state, after the SEL

    def allows_sending(self) -> bool:
        """Whether ports on this operator can send."""
        return self in {Operator.NONE, Operator.O_I, Operator.O_F}

    def allows_receiving(self) -> bool:
        """Whether ports on this operator can receive."""
        return self in {Operator.NONE, Operator.F_INIT, Operator.S}


class Port:
    """A port on a component.

    Ports are used by component to send or receive messages on. They are
    connected by conduits to enable communication between components.

    Attributes:
        name: The name of the port.
        operator: The MMSL operator in which this port is used.

    """

    def __init__(self, name: Identifier, operator: Operator) -> None:
        """Create a Port.

        Args:
            name: The name of the port.
            operator: The MMSL Operator in which this port is used.
        """
        self.name = name  # type: Identifier
        self.operator = operator  # type: Operator


class Ports:
    """Ports declaration for a component.

    Ports objects compare for equality by value. The names may be
    specified as a list of strings, or separated by spaces in a single
    string. If a particular operator has no associated ports, it may
    be omitted. For example:

    .. code:: yaml

        ports:
          f_init:   # list of names
          - a
          - b
          o_i: c d  # on one line, space-separated
          s: e      # single port
                    # o_f omitted as it has no ports

    Attributes:
        f_init: The ports associated with the F_INIT operator.
        o_i: The ports associated with the O_I operator
        s: The ports associated with the S operator.
        o_f: The ports associated with the O_F operator
    """
    def __init__(
            self, f_init: Union[None, str, List[str]] = None,
            o_i: Union[None, str, List[str]] = None,
            s: Union[None, str, List[str]] = None,
            o_f: Union[None, str, List[str]] = None) -> None:
        """Create a Ports declaration.

        Args:
            f_init: The ports associated with the F_INIT operator.
            o_i: The ports associated with the O_I operator
            s: The ports associated with the S operator.
            o_f: The ports associated with the O_F operator
        """
        def to_list(ports: Union[None, str, List[str]]) -> List[Identifier]:
            if ports is None:
                return list()

            if isinstance(ports, str):
                ports = ports.split(' ')
            return list(map(Identifier, ports))

        self.f_init = to_list(f_init)
        self.o_i = to_list(o_i)
        self.s = to_list(s)
        self.o_f = to_list(o_f)

        all_names = sorted(self.port_names())
        for i in range(len(all_names) - 1):
            if all_names[i] == all_names[i+1]:
                raise RuntimeError(
                        'Invalid ports specification: port "{}" is specified'
                        ' more than once. Port names must be unique within'
                        ' the component.')

    def port_names(self) -> Iterable[Identifier]:
        """Returns an iterable containing the names of all ports."""
        return self.f_init + self.o_i + self.s + self.o_f

    def all_ports(self) -> Iterable[Port]:
        """Returns an iterable containing all ports."""
        return (
                [Port(n, Operator.F_INIT) for n in self.f_init] +
                [Port(name, Operator.O_I) for name in self.o_i] +
                [Port(name, Operator.S) for name in self.s] +
                [Port(name, Operator.O_F) for name in self.o_f])

    def operator(self, port_name: Identifier) -> Operator:
        """Looks up the operator for a given port.

        Args:
            port_name: Name of the port to look up.

        Return:
            The operator for that port.

        Raises:
            KeyError: If no port with this name was found.
        """
        if port_name in self.f_init:
            return Operator.F_INIT

        if port_name in self.o_i:
            return Operator.O_I

        if port_name in self.s:
            return Operator.S

        if port_name in self.o_f:
            return Operator.O_F

        raise KeyError('No port named "{}" was found'.format(port_name))

    _yatiml_defaults = {
            'f_init': [],
            'o_i': [],
            's': [],
            'o_f': []}  # type: Dict[str, Optional[List[str]]]

    @classmethod
    def _yatiml_sweeten(cls, node: yatiml.Node) -> None:
        node.remove_attributes_with_default_values(cls)


class Component:
    """An object declaring a simulation component.

    Simulation components are things like submodels, scale bridges,
    proxies, and any other program that makes up a model. This class
    represents a declaration of a set of instances of a simulation
    component, and it's used to describe which instances are needed to
    perform a certain simulation.

    Attributes:
        name (ymmsl.Reference): The name of this component.
        implementation (ymmsl.Reference): A reference to the
                implementation to use.
        multiplicity (List[int]): The shape of the array of instances
                that execute simultaneously.
        ports (Optional[Ports]): The ports of this component,
                organised by operator. None if not specified.

    """

    def __init__(self, name: str, implementation: Optional[str] = None,
                 multiplicity: Union[None, int, List[int]] = None,
                 ports: Optional[Ports] = None) -> None:
        """Create a Component.

        Args:
            name: The name of the component; must be a valid
                    Reference.
            implementation: The name of the implementation; must be a
                    valid Reference.
            multiplicity: A list of ints describing the shape of the
                    set of instances.
            ports: The ports used by this component to communicate,
                    organised by operator.

        """
        self.name = Reference(name)
        if implementation is None:
            self.implementation = None      # type: Optional[Reference]
        else:
            self.implementation = Reference(implementation)
            for part in self.implementation:
                if isinstance(part, int):
                    raise ValueError('Component implementation {} contains a'
                                     ' subscript, which is not'
                                     ' allowed.'.format(self.name))

        if multiplicity is None:
            self.multiplicity = list()
        elif isinstance(multiplicity, int):
            self.multiplicity = [multiplicity]
        else:
            self.multiplicity = multiplicity

        self.ports = ports

    def __str__(self) -> str:
        """Returns a string representation of the object."""
        result = str(self.name)
        for dim in self.multiplicity:
            result += '[0:{}]'.format(dim)
        return result

    def __repr__(self) -> str:
        """Returns a string representation of the object."""
        return 'Component({})'.format(self.name)

    def instances(self) -> List[Reference]:
        """Creates a list of instances needed.

        Returns:
            A list with one Reference for each instance of this
            component.
        """
        def increment(index: List[int], dims: List[int]) -> None:
            # assumes index and dims are the same length > 0
            # modifies index argument
            i = len(index) - 1
            index[i] += 1
            while index[i] == dims[i]:
                index[i] = 0
                i -= 1
                if i == -1:
                    break
                index[i] += 1

        def generate_indices(multiplicity: List[int]) -> List[List[int]]:
            # n-dimensional counter
            indices = list()    # type: List[List[int]]

            index = [0] * len(multiplicity)
            indices.append(index.copy())
            increment(index, multiplicity)
            while sum(index) > 0:
                indices.append(index.copy())
                increment(index, multiplicity)
            return indices

        result = list()     # type: List[Reference]

        if len(self.multiplicity) == 0:
            result.append(self.name)
        else:
            for index in generate_indices(self.multiplicity):
                result.append(self.name + index)

        return result

    def _yatiml_attributes(self) -> OrderedDict:
        """Reorder attributes for saving to YAML."""
        return OrderedDict([
            ('name', self.name),
            ('ports', self.ports),
            ('multiplicity', self.multiplicity),
            ('implementation', self.implementation)])

    @classmethod
    def _yatiml_recognize(cls, node: yatiml.UnknownNode) -> None:
        node.require_attribute('name', str)

    @classmethod
    def _yatiml_savorize(cls, node: yatiml.Node) -> None:
        if node.has_attribute('multiplicity'):
            if node.has_attribute_type('multiplicity', int):
                attr = node.get_attribute('multiplicity')
                start_mark = attr.yaml_node.start_mark
                end_mark = attr.yaml_node.end_mark
                new_seq = yaml.nodes.SequenceNode(
                        'tag:yaml.org,2002:seq', [attr.yaml_node], start_mark,
                        end_mark)
                node.set_attribute('multiplicity', new_seq)

    @classmethod
    def _yatiml_sweeten(cls, node: yatiml.Node) -> None:
        multiplicity = node.get_attribute('multiplicity')
        items = multiplicity.seq_items()
        if len(items) == 0:
            node.remove_attribute('multiplicity')
        elif len(items) == 1:
            node.set_attribute('multiplicity', items[0].get_value())

        ports_node = node.get_attribute('ports').yaml_node
        if len(ports_node.value) == 0:
            node.remove_attribute('ports')



"""Defines the YAML document and version tag."""
import yatiml


class Document:
    """Represents a yMMSL document.

    This gets mixed in with a top-level content class, and takes
    care of a special ymmsl_version attribute in the root of the
    YAML file.
    """

    @classmethod
    def _yatiml_recognize(cls, node: yatiml.UnknownNode) -> None:
        node.require_mapping()
        node.require_attribute('ymmsl_version')
        node.require_attribute_value('ymmsl_version', 'v0.1')

    @classmethod
    def _yatiml_sweeten(cls, node: yatiml.Node) -> None:
        node.set_attribute('ymmsl_version', 'v0.1')
        # The above adds the attribute to the end, but we want it at
        # the top; this moves it there.
        node.yaml_node.value.insert(0, node.yaml_node.value[-1])
        del node.yaml_node.value[-1]

    @classmethod
    def _yatiml_savorize(cls, node: yatiml.Node) -> None:
        node.remove_attribute('ymmsl_version')



"""Python bindings for yMMSL.

This package contains all the classes needed to represent a yMMSL file,
as well as to read and write yMMSL files.
"""

from ymmsl.checkpoint import (
        CheckpointRule, CheckpointRangeRule, CheckpointAtRule, Checkpoints)
from ymmsl.component import Component, Operator, Port, Ports
from ymmsl.configuration import Configuration, PartialConfiguration
from ymmsl.execution import (
        ExecutionModel, Implementation, MPICoresResReq, MPINodesResReq,
        ResourceRequirements, ThreadedResReq, KeepsStateForNextUse)
from ymmsl.settings import Settings, SettingValue
from ymmsl.identity import Identifier, Reference
from ymmsl.io import load, dump, save
from ymmsl.model import Conduit, Model, ModelReference


__version__ = '0.13.0'
__author__ = 'Lourens Veen'
__email__ = 'l.veen@esciencecenter.nl'


__all__ = [
        'CheckpointRule', 'CheckpointRangeRule', 'CheckpointAtRule',
        'Checkpoints', 'Component', 'Conduit', 'Configuration', 'dump',
        'ExecutionModel', 'Identifier', 'Implementation',
        'KeepsStateForNextUse', 'load', 'Model', 'ModelReference',
        'MPICoresResReq', 'MPINodesResReq', 'Operator', 'PartialConfiguration',
        'Port', 'Ports', 'Reference', 'ResourceRequirements', 'save',
        'Settings', 'SettingValue', 'ThreadedResReq']



"""Definitions for specifying model settings."""
from collections import OrderedDict
from collections.abc import MutableMapping
from copy import deepcopy
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import yatiml

from ymmsl.identity import Reference


SettingValue = Union[
        str, int, float, bool, List[float], List[List[float]],
        yatiml.bool_union_fix]


class Settings(MutableMapping):
    """Settings for doing an experiment.

    An experiment is done by running a model with particular settings, \
    for the submodel scales, model parameters and any other configuration.
    """

    def __init__(
            self,
            settings: Optional[Dict[str, SettingValue]] = None
            ) -> None:
        """Create a Settings object.

        This will make a deep copy of the settings argument, if
        given.

        Args:
            settings: Setting values to initialise a model with.

        """
        self._store = OrderedDict()  # type: Dict[Reference, SettingValue]

        if settings is not None:
            for key, value in settings.items():
                self[key] = deepcopy(value)

    def __eq__(self, other: Any) -> bool:
        """Returns whether keys and values are identical."""
        if not isinstance(other, Settings):
            return NotImplemented
        return self._store == other._store

    def __str__(self) -> str:
        """Represent as a string."""
        return str(self.as_ordered_dict())

    def __getitem__(self, key: Union[str, Reference]) -> SettingValue:
        """Returns an item, implements settings[name]."""
        if isinstance(key, str):
            key = Reference(key)
        return self._store[key]

    def __setitem__(self, key: Union[str, Reference], value: SettingValue
                    ) -> None:
        """Sets a value, implements settings[name] = value."""
        if isinstance(key, str):
            key = Reference(key)
        self._store[key] = value

    def __delitem__(self, key: Union[str, Reference]) -> None:
        """Deletes a value, implements del(settings[name])."""
        if isinstance(key, str):
            key = Reference(key)
        del self._store[key]

    def __iter__(self) -> Iterator[Tuple[Reference, SettingValue]]:
        """Iterate through the settings' key, value pairs."""
        return iter(self._store)  # type: ignore

    def __len__(self) -> int:
        """Returns the number of settings."""
        return len(self._store)

    def ordered_items(self) -> List[Tuple[Reference, SettingValue]]:
        """Return settings as a list of tuples."""
        result = list()
        for key, value in self._store.items():
            result.append((key, value))
        return result

    def copy(self) -> 'Settings':
        """Makes a shallow copy of these settings and returns it."""
        new_settings = Settings()
        new_settings._store = self._store.copy()
        return new_settings

    def as_ordered_dict(self) -> OrderedDict:
        """Represent as a dictionary of plain built-in types.

        Returns: A dictionary that uses only built-in types, containing
            the configuration.
        """
        odict = OrderedDict()     # type: OrderedDict[str, SettingValue]
        for key, value in self._store.items():
            odict[str(key)] = value
        return odict

    @classmethod
    def _yatiml_recognize(cls, node: yatiml.UnknownNode) -> None:
        # In the YAML file, a Settings is just a mapping...
        node.require_mapping()

    def _yatiml_attributes(self) -> Dict:
        # ...so we just give YAtiML our internal mapping to serialise
        return self._store

    @classmethod
    def _yatiml_savorize(cls, node: yatiml.Node) -> None:
        # wrap the existing mapping into a new mapping with attribute settings
        setting_values = node.yaml_node
        node.make_mapping()
        node.set_attribute('settings', setting_values)

    @classmethod
    def _yatiml_sweeten(cls, node: yatiml.Node) -> None:
        # format lists and arrays nicely
        for _, value_node in node.yaml_node.value:
            if value_node.tag == 'tag:yaml.org,2002:seq':
                # this attribute is a list or list-of-list
                if len(value_node.value) > 0:
                    if value_node.value[0].tag != 'tag:yaml.org,2002:seq':
                        value_node.flow_style = True
                    else:
                        value_node.flow_style = False
                        for row_node in value_node.value:
                            row_node.flow_style = True
